/*
 * Network Interface Monitor via ParPort LEDs, light bulbs, etc.
 * Copyright (C) 1999. Vlatko Kosturjak 
 * Uses some code from Ifled by Mattias Wadman <napolium@sudac.org>
 * Distributed under GPL.
 */

/*#define USELPDIFF 1*/

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <linux/kd.h>
#include <sys/utsname.h>
#include <signal.h>
#include <unistd.h>

#ifdef USELPDIFF
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/lp.h>
#endif

#define LINESIZE 255

#define DEFLPTADR 0x378
#define DEFLPDEV "/dev/lp0"
#define DEFIFDEV "lo"

#define TRUE	1
#define FALSE	0

#define IF_RX		0
#define IF_TX		1
#define IF_COLL		2
#define IF_UNDEF	3
#define IF_DROP_RX 	4
#define IF_DROP_TX	5
#define IF_ERR_TX	6
#define IF_ERR_RX	7

#define IF_RXTX		8
#define IF_DROP		9
#define IF_ERR		10
#define IF_NONE		11

#define OPT_FORK	1
#define OPT_KERNEL_2_0	2
#define OPT_INVERT	4
#define OPT_ALTKBCODE	8

unsigned long int if_info[8]; 
unsigned long int l_if_info[8]; 
unsigned char changed[8];
char options = 0;
int LPTport = DEFLPTADR;
char LPdev[20] = DEFLPDEV;
char IFdev[20] = DEFIFDEV;
int fd;
int debugprint=0, dummymode=0, daemonmode=0;

__inline__ static void outp(unsigned short port,char value)
{  __asm__ volatile ("outb %0,%1"
            ::"a" ((char) value), "d"((unsigned short) port));
}

void update_netproc(char *interface)
{
	char line[LINESIZE];
	char *tok;
	char temp;
	FILE *procfd;
	if((procfd = fopen("/proc/net/dev","r")) == NULL)
	{
		perror("Unable to open /proc/net/dev.");	
		exit (1);
	}
	fgets(line,LINESIZE,procfd);
	while(!feof(procfd))
	{
		tok = strtok(line,":");
		if(strstr(tok,interface))
		{
			if(options & OPT_KERNEL_2_0)
			{
				for(temp=0;temp < 10;temp++)
				{	
					tok = strtok(NULL," ");
					if(temp == 0)
						if_info[IF_RX] = atol(tok);
					else if(temp == 1)
						if_info[IF_ERR_RX] = atol(tok);
					else if(temp == 2)
						if_info[IF_DROP_RX] = atol(tok);
					else if(temp == 5)
						if_info[IF_TX] = atol(tok);
					else if(temp == 6)
						if_info[IF_ERR_TX] = atol(tok);
					else if(temp == 7)
						if_info[IF_DROP_TX] = atol(tok);
					else if(temp == 9)
						if_info[IF_COLL] = atol(tok);
				}
			}
			else
			{
				for(temp=0;temp < 14;temp++)
				{	
					tok = strtok(NULL," ");
					if(temp == 0)
						if_info[IF_RX] = atol(tok);
					else if(temp == 2)
						if_info[IF_ERR_RX] = atol(tok);
					else if(temp == 3)
						if_info[IF_DROP_RX] = atol(tok);
					else if(temp == 8)
						if_info[IF_TX] = atol(tok);
					else if(temp == 10)
						if_info[IF_ERR_TX] = atol(tok);
					else if(temp == 11)
						if_info[IF_DROP_TX] = atol(tok);
					else if(temp == 13)
						if_info[IF_COLL] = atol(tok);
				}
			}
			fclose(procfd);
			return;
		}
		fgets(line,LINESIZE,procfd);
	}
	fclose(procfd);
	fprintf(stderr,"Unable to find interface.");
	exit (1);
}

char is_changed()
{
	if (if_info[IF_RX] == l_if_info[IF_RX])
	if (if_info[IF_TX] == l_if_info[IF_TX])
	if (if_info[IF_DROP_RX] == l_if_info[IF_DROP_RX])
	if (if_info[IF_DROP_TX] == l_if_info[IF_DROP_TX])
	if (if_info[IF_ERR_RX] == l_if_info[IF_ERR_RX])
	if (if_info[IF_ERR_TX] == l_if_info[IF_ERR_TX])
	if (if_info[IF_COLL] == l_if_info[IF_COLL])
	return FALSE;

	return TRUE;
}

void printvalues (void)
{
	printf ("\rRX:%u TX:%u DR:%u DT:%u ER:%u ET:%u CO:%u",changed[IF_RX],changed[IF_TX],changed[IF_DROP_RX],changed[IF_DROP_TX],changed[IF_ERR_RX],changed[IF_ERR_TX],changed[IF_COLL]);
} 


unsigned char makeit()
{
	int lptvalue=0;
	int conditions[8]={1,2,4,8,16,32,64,128};
	int i,ret;

	changed[IF_UNDEF]=1;	

	if (if_info[IF_RX] == l_if_info[IF_RX]) changed[IF_RX]=1;
	else changed[IF_RX]=0;

	if (if_info[IF_TX] == l_if_info[IF_TX]) changed[IF_TX]=1;
	else changed[IF_TX]=0;

	if (if_info[IF_DROP_RX] == l_if_info[IF_DROP_RX]) changed[IF_DROP_RX]=1;
	else changed[IF_DROP_RX]=0;

	if (if_info[IF_DROP_TX] == l_if_info[IF_DROP_TX]) changed[IF_DROP_TX]=1;
	else changed[IF_DROP_TX]=0;

	if (if_info[IF_ERR_RX] == l_if_info[IF_ERR_RX]) changed[IF_ERR_RX]=1;
	else changed[IF_DROP_RX]=0;

	if (if_info[IF_ERR_TX] == l_if_info[IF_ERR_TX]) changed[IF_ERR_TX]=1;
	else changed[IF_ERR_TX]=0;

	if (if_info[IF_COLL] == l_if_info[IF_COLL]) changed[IF_COLL]=1;
	else changed[IF_COLL]=0;
	if (dummymode==0)
	{
		for (i=0;i<8;i++)
		{
			if (changed[i]==0) lptvalue+=conditions[i];
		}
#ifdef USELPDIFF
		ret = ioctl (fd, LPWRITERAW, &lptvalue);
		if (ret < 0) 
		{
			perror ("lpdiff ioctl");
			exit (11);
		}
#else
		outp (LPTport,lptvalue);
#endif
	}
	if (debugprint==1) printvalues();
}

void prepare (void)
{
	if (dummymode==0)
	{
	#ifdef USELPDIFF
	fd=open(LPdev,O_RDWR | O_NONBLOCK);
	if (fd < 0)
	{
		perror ("lpdiff open");
		exit (10);
	}
	#else
	if (ioperm(LPTport,3,1))
	{
		perror ("lpt ioperm");
		exit (1);
	}
	#endif
	}
}

void finish (void)
{
	if (dummymode==0)
	{
	#ifdef USELPDIFF
	close(fd);
	#else
	if (ioperm(LPTport,3,0))
	{
		perror("lpt ioperm");
		exit (1);
	}
	#endif
	}
}

void intro (void)
{
	printf ("Network Interface Monitor via Parport LEDs v0.2. Copyright (C) 1999. KoSt.\n\n");
}

void printhelp (void)
{
const char *help =
	"Usage: parport -p lptadr -d device -i iface -l delay -b -e -g\n"
	"\t-b daemon mode\n"
	"\t-e emulate sending to LPT\n"
	"\t-g debug printout\n\n";

printf ("%s",help);
}

void fork_program()
{
	pid_t program_pid;
	program_pid = fork();
	if(program_pid == -1)
	{	
		perror("Unable to fork program.");
		exit(22);
	}
	if(program_pid != 0)
		exit(0);
}

int main(int argc, char *argv[])
{
	int delay = 50;
	struct utsname uname_dummy;
	char arg_dummy;
	
	setvbuf (stdout,NULL,_IONBF,0); /* turn off buffering */

	for(arg_dummy=1;arg_dummy < argc;arg_dummy++)
	{
	if(argv[arg_dummy][0] != '-')
	{
		printf("Error: option: %s\n",argv[arg_dummy]);
		exit(1);
	}
	switch(argv[arg_dummy][1])
	{
		case 'h':
			printhelp();
			exit (0);
			break;
		case 'b':
			daemonmode=1;
			break;
		case 'g':
			debugprint=1;
			break;
		case 'e':
			dummymode=1;
			break;
		case 'l':
			if(argv[arg_dummy+1] == NULL)
			{	
				fprintf(stderr,"-l option needs an integer");
				exit(2);
			}
			delay = atol(argv[arg_dummy+1]);
			arg_dummy++;
			break;
		case 'i':
			if(argv[arg_dummy+1] == NULL)
			{
				fprintf(stderr,"-i option needs an interface (e.g. eth0)");
				exit(2);
			}
			strncpy(IFdev,argv[arg_dummy+1],20);
			arg_dummy++;
			break;
		case 'd':
			if(argv[arg_dummy+1] == NULL)
			{
				fprintf(stderr,"-d option needs a device (e.g. /dev/lp0)");
				exit(2);
			}
			strncpy(LPdev,argv[arg_dummy+1],20);
			arg_dummy++;
			break;
		case 'p':
			if(argv[arg_dummy+1] == NULL)
			{	
				fprintf(stderr,"-p option needs a port address (e.g. 378)");
				exit(2);
			}
			LPTport = atoi(argv[arg_dummy+1]);
			arg_dummy++;
			break;
		default:
			printf("Error: option: %s\n",argv[arg_dummy]);
			exit(1);
			break;
	}
	}

	if (daemonmode==1) fork_program();    /* fork into background or... */
	else intro(); /* display greeting */

	uname(&uname_dummy);
	if(strncmp(uname_dummy.release,"2.0",3) == 0)
		options |= OPT_KERNEL_2_0;
	prepare();
	update_netproc(IFdev);
	memcpy(&l_if_info,&if_info,sizeof(if_info));
	while(1)
	{
		update_netproc(IFdev);
		makeit();
		memcpy(&l_if_info,&if_info,sizeof(if_info));
		usleep(delay*1000);
	}
	finish();
	printf ("\nEnd.\n");
}
